# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import re
import sys

sys.path.insert(0, os.path.abspath(".."))


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "s1etad"
copyright = "2020-2025, s1etad Developers"
author = "Nuno Miranda, Antonio Valentino"

with open("../s1etad/__init__.py") as fd:
    s = fd.read()

pattern = (
    r"^__version__( )?=( )?"
    r'(?P<q>[\'"])(?P<r>(?P<v>\d+\.\d+(\.\d+)?).*)(?P=q)'
)

# The short X.Y version
version = re.search(pattern, s, re.M).group("v")

# The full version, including alpha/beta/rc tags
release = re.search(pattern, s, re.M).group("r")


# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    # "sphinx.ext.autosectionlabel",
    # "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.duration",
    "sphinx.ext.extlinks",
    # "sphinx.ext.githubpages",
    # "sphinx.ext.graphviz",
    "sphinx.ext.ifconfig",
    # "sphinx.ext.imgconverter",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    # "sphinx.ext.linkcode",  # needs_sphinx = "1.2"
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    # "sphinx.ext.imgmath",
    # "sphinx.ext.jsmath",
    "sphinx.ext.mathjax",
    "nbsphinx",
    "sphinx_rtd_theme",
]

try:
    import sphinxcontrib.spelling  # noqa: F401
except ImportError:
    print("sphinxcontrib.spelling extensiion not available")
else:
    extensions.append("sphinxcontrib.spelling")

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_theme_options = {
    # 'vcs_pageview_mode': 'blob',
}
html_context = {
    # "github_url": "https://github.com/s1tools/s1-etad",
    "display_github": True,
    "github_user": "s1tools",
    "github_repo": "s1-etad",
    "github_version": "main",
    "conf_py_path": "/docs/",
}

# html_last_updated_fmt = ""


# -- Extension configuration -------------------------------------------------

# -- Options for autodoc extension -------------------------------------------
# autoclass_content = 'both'
autodoc_member_order = "groupwise"
# autodoc_default_options = {
#     "members": True,
#     "undoc-members": True,
#     "show-inheritance": True,
# }
autodoc_mock_imports = [
    "numpy",
    "scipy",
    "lxml",
    "netCDF4",
    "pandas",
    "dateutil",
    "shapely",
    "simplekml",
    # "osgeo",
    "matplotlib",
]


# -- Options for autosummary extension ---------------------------------------
autosummary_generate = True
# autosummary_mock_imports = []
# autosummary_ignore_module_all = False


# -- Options for intersphinx extension ---------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}


# -- Options for extlinks extension ------------------------------------------

extlinks = {
    "issue": ("https://github.com/s1tools/s1-etad/issues/%s", "gh-%s"),
}


# -- Options for todo extension ----------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/extensions/todo.html#configuration

todo_include_todos = True


# nbsphinx_execute = "always"  # "nener", "always", "auto"



"""pytest fixtures for s1-etad testing."""

import pathlib

import pooch
import pytest

import s1etad

from . import dataset

S1_ETAD_IW_NAME = "s1_setap_3-0-0_2025-05-30_etad_iw-tar"
S1_ETAD_SM_NAME = "s1_setap_3-0-0_2025-05-30_etad_sm-tar"


def _get_product_path(name: str) -> pathlib.Path:
    files = dataset.datarepo.fetch(name, pooch.Untar())
    assert files
    return pathlib.Path(files[0]).parent


@pytest.fixture
def sm_etad_filename() -> pathlib.Path:
    return _get_product_path(S1_ETAD_SM_NAME)


@pytest.fixture(scope="session")
def sm_etad_product():
    filename = _get_product_path(S1_ETAD_SM_NAME)
    return s1etad.Sentinel1Etad(filename)


@pytest.fixture(scope="session")
def sm_etad_swath(sm_etad_product):
    swath_id = sm_etad_product.swath_list[0]
    return sm_etad_product[swath_id]


@pytest.fixture(scope="session")
def sm_etad_burst(sm_etad_swath):
    burst_id = sm_etad_swath.burst_list[0]
    return sm_etad_swath[burst_id]


@pytest.fixture
def iw_etad_filename() -> pathlib.Path:
    return _get_product_path(S1_ETAD_IW_NAME)


@pytest.fixture(scope="session")
def iw_etad_product():
    filename = _get_product_path(S1_ETAD_IW_NAME)
    return s1etad.Sentinel1Etad(filename)


@pytest.fixture(scope="session")
def iw_etad_swath(iw_etad_product):
    swath_id = iw_etad_product.swath_list[0]
    return iw_etad_product[swath_id]


@pytest.fixture(scope="session")
def iw_etad_burst(iw_etad_swath):
    burst_id = iw_etad_swath.burst_list[0]
    return iw_etad_swath[burst_id]



"""S1-ETAD data model core classes."""

import enum
import pathlib
import weakref
import datetime
import warnings
import itertools
import collections
from typing import Literal

import numpy as np
import pandas as pd
import shapely.ops
from lxml import etree
from scipy import constants
from netCDF4 import Dataset
from shapely.geometry import Polygon, MultiPolygon
from shapely.geometry.base import BaseGeometry

from ._s1utils import Sentinel1ProductName

__all__ = [
    "Sentinel1Etad",
    "Sentinel1EtadSwath",
    "Sentinel1EtadBurst",
    "ECorrectionType",
]


class ECorrectionType(enum.Enum):
    """Enumeration for correction types."""

    TROPOSPHERIC = "tropospheric"
    TROPOSPHERIC_GRADIENT = "tropospheric_gradient"
    IONOSPHERIC = "ionospheric"
    GEODETIC = "geodetic"
    OTL = "otl"
    BISTATIC = "bistatic"
    DOPPLER = "doppler"
    FMRATE = "fmrate"
    SUM = "sum"


CorrectionType = ECorrectionType | str


_CORRECTION_NAMES_MAP = {
    "tropospheric": {"x": "troposphericCorrectionRg"},
    "tropospheric_gradient": {"x": "troposphericCorrectionHeightGradientRg"},
    "ionospheric": {"x": "ionosphericCorrectionRg"},
    "geodetic": {"x": "geodeticCorrectionRg", "y": "geodeticCorrectionAz"},
    "otl": {
        "x": "oceanTidalLoadingCorrectionRg",
        "y": "oceanTidalLoadingCorrectionAz",
    },
    "bistatic": {"y": "bistaticCorrectionAz"},
    "doppler": {"x": "dopplerRangeShiftRg"},
    "fmrate": {"y": "fmMismatchCorrectionAz"},
    "sum": {"x": "sumOfCorrectionsRg", "y": "sumOfCorrectionsAz"},
}


_STATS_TAG_MAP = {
    ECorrectionType.TROPOSPHERIC: "troposphericCorrection",
    ECorrectionType.TROPOSPHERIC_GRADIENT: (
        "troposphericCorrectionHeightGradient"
    ),
    ECorrectionType.IONOSPHERIC: "ionosphericCorrection",
    ECorrectionType.GEODETIC: "geodeticCorrection",
    ECorrectionType.OTL: "oceanTidalLoadingCorrection",
    ECorrectionType.BISTATIC: "bistaticCorrection",
    ECorrectionType.DOPPLER: "dopplerRangeShift",
    ECorrectionType.FMRATE: "fmMismatchCorrection",
    ECorrectionType.SUM: "sumOfCorrections",
}


Statistics = collections.namedtuple("Statistics", ["min", "mean", "max"])


class Sentinel1Etad:
    """Sentinel-1 ETAD product.

    Class to decode and access the elements of the Sentinel ETAD product
    which specification is governed by ETAD-DLR-PS-0014.

    The indexing operator '[] '(implemented with the __getitem__ method)
    returns a Sentinel1EtadSwath instance.

    Parameters
    ----------
    product : str or pathlib.Path
        path of the S1-ETAD product (it is a directory)

    Attributes
    ----------
    product : pathlib.Path
        path of the S1-ETAD product (it is a directory)
    burst_catalogue : pandas.DataFrame
        dataframe containing main information of all bursts present in
        the product
    ds : netCDF.Dataset
        (provisional) the NetCDF.Dataset in which data are stored

    """

    def __init__(self, product):
        """Initialize a `Sentinel1Etad` object."""
        # TODO: make this read-only (property)
        self.product = pathlib.Path(product)
        if not self.product.is_dir():
            raise FileNotFoundError(str(product))
        # TODO: ds should not be exposed
        self.ds = self._init_measurement_dataset()
        self._annot = self._init_annotation_dataset()
        self._processor_version = self._init_processor_version()
        self.burst_catalogue = self._init_burst_catalogue()
        self._swaths = {}

    def _init_measurement_dataset(self):
        """Open the nc dataset."""
        # @TODO: retrieve form manifest
        netcdf_files = list(self.product.glob("measurement/*.nc"))
        if not netcdf_files:
            raise FileNotFoundError(
                "Unable to find NetCDF files in "
                f"{self.product.joinpath('measurement')}"
            )
        assert len(netcdf_files) == 1
        netcdf_file = netcdf_files[0]
        rootgrp = Dataset(netcdf_file, "r")
        rootgrp.set_auto_mask(False)
        return rootgrp

    def _init_annotation_dataset(self):
        """Open the xml annotation dataset."""
        # @TODO: retrieve form manifest
        xml_files = list(self.product.glob("annotation/*.xml"))
        if not xml_files:
            raise FileNotFoundError(
                "Unable to find XML files in "
                f"{self.product.joinpath('annotation')}"
            )
        assert len(xml_files) == 1
        xml_file = str(xml_files[0])
        root = etree.parse(xml_file).getroot()
        return root

    def _init_processor_version(self) -> str:
        processor_version = self._annot.findtext(".//processorVersion")
        if processor_version is None:
            raise RuntimeError("unable to detect the processor version")
        return processor_version

    def __getitem__(self, index):
        assert index in self.swath_list, f"{index} is not in {self.swath_list}"
        if index not in self._swaths:
            self._swaths[index] = Sentinel1EtadSwath(self.ds[index], self)
        return self._swaths[index]

    def __iter__(self):
        yield from self.iter_swaths()

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.product}")  # 0x{id(self):x}'

    def __str__(self):
        return f'{self.__class__.__name__}("{self.product.name}")'

    @property
    def number_of_swath(self):
        """Return the number of swaths in the product."""
        return len(self.ds.groups)

    @property
    def swath_list(self):
        """Return the list of swath identifiers (str) in the product."""
        return list(self.ds.groups.keys())

    def s1_product_list(self):
        """Return the list of S-1 products used to compose the ETAD one."""
        df = self.burst_catalogue

        # this ensures that each product name is located at the correct pIndex
        product_list = [
            item[1]
            for item in sorted(
                set(zip(df["pIndex"], df["productID"], strict=True))
            )
        ]

        return product_list

    @property
    def grid_spacing(self):
        """Return the grid spacing in meters."""
        xp_list = {
            "x": ".//correctionGridRangeSampling",
            "y": ".//correctionGridAzimuthSampling",
        }
        dd = {}
        for tag, xp in xp_list.items():
            dd[tag] = self._xpath_to_list(self._annot, xp, dtype=float)
        dd["unit"] = "m"
        return dd

    @property
    def grid_sampling(self):
        """Return the grid spacing in s."""
        xp_list = {
            "x": ".//productInformation/gridSampling/range",
            "y": ".//productInformation/gridSampling/azimuth",
        }
        dd = {}
        for tag, xp in xp_list.items():
            dd[tag] = self._xpath_to_list(self._annot, xp, dtype=float)
        dd["unit"] = "s"
        return dd

    @property
    def min_azimuth_time(self):
        """Return the minimum azimuth time of all bursts in the product."""
        return datetime.datetime.fromisoformat(self.ds.azimuthTimeMin)

    @property
    def max_azimuth_time(self):
        """Return the maximum azimuth time of all bursts in the product."""
        return datetime.datetime.fromisoformat(self.ds.azimuthTimeMax)

    @property
    def min_range_time(self):
        """Return the minimum range time of all bursts in the product."""
        return self.ds.rangeTimeMin

    @property
    def max_range_time(self):
        """Return the maximum range time of all bursts in the product."""
        return self.ds.rangeTimeMax

    @property
    def vg(self):
        """Mean ground velocity [m/s]."""
        try:
            xp = (
                "productInformation/gridGroundSampling/"
                "averageZeroDopplerVelocity"
            )
            vg = float(self._annot.find(xp).taxt)
        except (AttributeError, ValueError):
            vg = self.grid_spacing["y"] / self.grid_sampling["y"]
        return vg

    def processing_setting(self):
        """Return the corrections performed.

        Read the xml file to identify the corrections performed.
        If a correction is not performed the matrix is filled with zeros.
        """
        correction_list = [
            "troposphericDelayCorrection",
            "troposphericDelayCorrectionGradient",
            "ionosphericDelayCorrection",
            "solidEarthTideCorrection",
            "oceanTidalLoadingCorrection",
            "bistaticAzimuthCorrection",
            "dopplerShiftRangeCorrection",
            "FMMismatchAzimuthCorrection",
        ]
        dd = {}
        xp_root = (
            "processingInformation/processor/setapConfigurationFile/"
            "processorSettings/"
        )
        for correction in correction_list:
            xp = xp_root + correction
            ret = self._xpath_to_list(self._annot, xp)
            if ret == "true":
                ret = True
            elif ret == "false":
                ret = False
            else:
                continue
            dd[correction] = ret
        return dd

    def _init_burst_catalogue(self):
        """Build the burst catalog.

        Using information stored in the NetCDF file create a
        pandas.DataFrame containing all the elements allowing to index
        properly a burst.
        """

        def _to_tdelta64(t):
            return np.float64(t * 1e9).astype("timedelta64[ns]")

        data = collections.defaultdict(list)
        t0 = np.datetime64(self.ds.azimuthTimeMin, "ns")
        for swath in self.ds.groups.values():
            for burst in swath.groups.values():
                ax = burst.variables["azimuth"]
                tmin = t0 + _to_tdelta64(ax[0])
                tmax = t0 + _to_tdelta64(ax[-1])

                data["bIndex"].append(burst.bIndex)
                data["pIndex"].append(burst.pIndex)
                data["sIndex"].append(burst.sIndex)
                data["productID"].append(burst.productID)
                data["swathID"].append(burst.swathID)
                data["azimuthTimeMin"].append(tmin)
                data["azimuthTimeMax"].append(tmax)

        df = pd.DataFrame(data=data)

        return df

    def query_burst(
        self,
        first_time=None,
        product_name=None,
        last_time=None,
        swath=None,
        geometry=None,
    ):
        """Query the burst catalogue to retrieve the burst matching by time.

        Parameters
        ----------
        first_time : datetime
            is set to None then set to the first time
        last_time : datetime
            if set to None the last_time = first_time
        product_name : str
            Name of a real S1 product e.g.
            S1B_IW_SLC__1SDV_20190805T162509_20190805T162...SAFE
        swath : str or list
            list of swathID e.g. 'IW1' or ['IW1'] or ['IW1', 'IW2']
        geometry : shapely.geometry.[Point, Polygon, ...]
            A shapely geometry for which intersection will be searched

        Returns
        -------
        pandas.DataFrame
            Filtered panda dataframe

        """
        # first sort the burst by time
        df = self.burst_catalogue.sort_values(by=["azimuthTimeMin"])
        if first_time is None:
            first_time = df.iloc[0].azimuthTimeMin
        if last_time is None:
            last_time = df.iloc[-1].azimuthTimeMax

        ix0 = (df.azimuthTimeMin >= first_time) & (
            df.azimuthTimeMax <= last_time
        )

        if product_name is not None:
            # build a regex based on the name to avoid issues with annotation
            # products and CRC
            product_name = Sentinel1ProductName(product_name)
            product_name.to_annotation(value="[AS]")
            product_name.crc = ""
            filter_ = product_name.recompose(with_suffix=False)
            ix0 = ix0 & self.burst_catalogue.productID.str.contains(
                filter_, regex=True
            )

        if swath is not None:
            if isinstance(swath, str):
                swath = [swath]
            ix0 = ix0 & df.swathID.isin(swath)

        if geometry is not None:
            bix_list = self.intersects(geometry)
            ix0 = ix0 & df.bIndex.isin(bix_list)

        return df.loc[ix0]

    def _selection_to_swath_list(self, selection=None):
        if selection is None:
            selection = self.burst_catalogue

        if isinstance(selection, pd.DataFrame):
            burst_selection = selection
            swath_list = selection.swathID.unique()
        elif isinstance(selection, str):
            burst_selection = None
            swath_list = [selection]
        else:
            # assume it is a list of swaths already
            import collections.abc

            assert isinstance(selection, collections.abc.Iterable)
            assert all(isinstance(item, str) for item in selection)
            burst_selection = None
            swath_list = selection

        return swath_list, burst_selection

    def iter_swaths(self, selection=None):
        """Iterate over swaths according to the specified selection.

        Parameters
        ----------
        selection : list(str) or pd.Dataframe, optional
            the list of selected swath IDs or the result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the swaths of the product.

        """
        swath_list, _ = self._selection_to_swath_list(selection)
        for swath_name in swath_list:
            yield self[swath_name]

    def iter_bursts(self, selection=None):
        """Iterate over burst according to the specified selection.

        Parameters
        ----------
        selection : list(int) or pd.Dataframe, optional
            the list of selected burst indexes or the result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the bursts of the product.

        """
        if selection is None:
            selection = self.burst_catalogue
        elif not isinstance(selection, pd.DataFrame):
            # assume it is a list of burst indexes
            bursts = selection
            if isinstance(bursts, int):
                bursts = [selection]
            # NOTE: preserve the order
            selection = self.burst_catalogue.bIndex.isin(bursts)

        assert isinstance(selection, pd.DataFrame)

        for _, row in selection.iterrows():
            burst = self[row.swathID][row.bIndex]
            yield burst

    @staticmethod
    def _xpath_to_list(
        root, xpath, dtype=None, namespace=None, parse_time_func=None
    ):
        ll = [elt.text for elt in root.findall(xpath, namespace)]
        if parse_time_func is not None:
            ll = [datetime.datetime.fromisoformat(t) for t in ll]
            ll = parse_time_func(ll)  # TODO: check
        arr = np.asarray(ll, dtype=dtype)

        if arr.size == 1:
            return arr.item(0)
        else:
            return arr

    def get_statistics(self, correction, meter: bool = False):
        """Return the global statistic value of the specified correction.

        The returned value is the pre-computed one that is stored in the
        XML annotation file of the product.

        Parameters
        ----------
        correction : str or ECorrectionType
            the corrections for which the statistic value is requested
        meter : bool
            if set to True then the returned value is expressed in meters,
            otherwise it is expressed in seconds (default: False)

        Returns
        -------
        dict
            a dictionary containing :class:`Statistics` (min, mean and max)
            for all available components of the specified correction:

            :x:
                a :class:`Statistics` instance relative to the range
                component of the specified correction
            :y:
                a :class:`Statistics` instance relative to the azimuth
                component of the specified correction
            :unit:
                the units of the returned statistics ("m" or "s")

        """
        stat_xp = "./qualityAndStatistics"
        target = ECorrectionType(correction)
        target_tag = _STATS_TAG_MAP[target]

        if target is ECorrectionType.TROPOSPHERIC_GRADIENT:
            units = "m/m" if meter else "s/m"
        else:
            units = "m" if meter else "s"

        statistics: dict[str, Statistics | str] = {}
        statistics["unit"] = units

        # NOTE: looping on element and heuristic test on tags is necessary
        #       due to inconsistent naming of range and azimuth element
        correction_elem = self._annot.find(f"{stat_xp}/{target_tag}")
        for elem in correction_elem:
            if "range" in elem.tag:
                direction = "x"
            elif "azimuth" in elem.tag:
                direction = "y"
            else:
                continue

            statistics[direction] = Statistics(
                float(elem.findtext(f'min[@unit="{units}"]')),
                float(elem.findtext(f'mean[@unit="{units}"]')),
                float(elem.findtext(f'max[@unit="{units}"]')),
            )

        if len(statistics) == 1:
            return {}

        return statistics

    def get_footprint(self, selection=None, merge: bool = False):
        """Return the footprints of all the bursts as MultiPolygon.

        It calls in the back the get_footprint of the Sentinel1EtadBurst class.

        Parameters
        ----------
        selection : list(str) or pd.Dataframe, optional
            the list of selected swath IDs or the result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the swaths of the product.
        merge : bool
            if set to True return a single polygon that is the union of the
            footprints of all bursts

        """
        poly_list = []
        swath_list, burst_selection = self._selection_to_swath_list(selection)
        for swath in self.iter_swaths(swath_list):
            poly_list.extend(swath.get_footprint(burst_selection).geoms)

        if merge:
            polys = shapely.ops.unary_union(poly_list)
        else:
            polys = MultiPolygon(poly_list)

        return polys

    def intersects(self, geometry: BaseGeometry):
        """Return the list of burst indexes intersecting the input geometry.

        Computes the intersection of the footprint of the swath (all bursts)
        with the input geometry.

        Parameters
        ----------
        geometry : shapely.geometry.[Point, Polygon, MultiPolygon, line]
            shapely geometry

        Returns
        -------
        list
            list of all the burst intersecting with the input shape geometry

        """
        lists_of_burst_indexes = [
            swath.intersects(geometry) for swath in self.iter_swaths()
        ]
        # return the flattened list
        return list(itertools.chain(*lists_of_burst_indexes))

    def _swath_merger(
        self,
        burst_var,
        selection=None,
        set_auto_mask: bool = False,
        meter: bool = False,
        fill_value: float = 0.0,
    ):
        if selection is None:
            df = self.burst_catalogue
        elif not isinstance(selection, pd.DataFrame):
            df = self.query_burst(swath=selection)
        else:
            assert isinstance(selection, pd.DataFrame)
            df = selection

        # NOTE: assume a specific order of swath IDs
        first_swath = self[df.swathID.min()]
        near_burst = first_swath[first_swath.burst_list[0]]
        last_swath = self[df.swathID.max()]
        far_burst = last_swath[last_swath.burst_list[0]]

        rg_first_time = near_burst.sampling_start["x"]
        rg_last_time = (
            far_burst.sampling_start["x"]
            + far_burst.sampling["x"] * far_burst.samples
        )
        az_first_time = df.azimuthTimeMin.min()
        az_last_time = df.azimuthTimeMax.max()
        az_ref_time = self.min_azimuth_time
        az_first_time_rel = (az_first_time - az_ref_time).total_seconds()

        sampling = self.grid_sampling
        dx = sampling["x"]
        dy = sampling["y"]

        num_samples = (
            np.round((rg_last_time - rg_first_time) / dx).astype(int) + 1
        )
        num_lines = (
            np.round(
                (az_last_time - az_first_time).total_seconds() / dy
            ).astype(int)
            + 1
        )

        img = np.full((num_lines, num_samples), fill_value=fill_value)
        # TODO: add some control option
        img = np.ma.array(img, mask=True, fill_value=fill_value)

        for swath in self.iter_swaths(df):
            # NOTE: use the private "Sentinel1EtadSwath._burst_merger" method
            # to be able to work only on the specified NetCDF variable
            dd_ = swath._burst_merger(
                burst_var,
                selection=df,
                set_auto_mask=set_auto_mask,
                meter=meter,
            )
            yoffset = dd_["first_azimuth_time"] - az_first_time_rel
            xoffset = dd_["first_slant_range_time"] - rg_first_time
            line_ofs = np.round(yoffset / dy).astype(int)
            sample_ofs = np.round(xoffset / dx).astype(int)

            slice_y = slice(line_ofs, line_ofs + dd_[burst_var].shape[0])
            slice_x = slice(sample_ofs, sample_ofs + dd_[burst_var].shape[1])

            img[slice_y, slice_x] = dd_[burst_var]

        return {
            burst_var: img,
            "first_azimuth_time": az_first_time,
            "first_slant_range_time": rg_first_time,
            "sampling": sampling,
        }

    def _core_merge_correction(
        self,
        prm_list,
        selection=None,
        set_auto_mask: bool = True,
        meter: bool = False,
    ):
        dd = {}
        for dim, field in prm_list.items():
            dd_ = self._swath_merger(
                field,
                selection=selection,
                set_auto_mask=set_auto_mask,
                meter=meter,
            )
            dd[dim] = dd_[field]
            dd["sampling"] = dd_["sampling"]
            dd["first_azimuth_time"] = dd_["first_azimuth_time"]
            dd["first_slant_range_time"] = dd_["first_slant_range_time"]

        dd["unit"] = "m" if meter else "s"

        # To compute lat/lon/h make a new selection with all gaps filled
        swath_list, _ = self._selection_to_swath_list(selection)
        near_swath = min(swath_list)
        far_swath = max(swath_list)
        idx = self.burst_catalogue.swathID >= near_swath
        idx &= self.burst_catalogue.swathID <= far_swath
        swaths = self.burst_catalogue.swathID[idx].unique()

        data = dd["x" if "x" in prm_list else "y"]
        lines = data.shape[0]
        duration = lines * self.grid_sampling["y"]
        duration = np.float64(duration * 1e9).astype("timedelta64[ns]")
        first_time = dd["first_azimuth_time"]
        last_time = first_time + duration

        filled_selection = self.query_burst(
            first_time=first_time, last_time=last_time, swath=swaths
        )

        dd["lats"] = self._swath_merger(
            "lats",
            selection=filled_selection,
            set_auto_mask=set_auto_mask,
            meter=False,
            fill_value=np.nan,
        )["lats"]
        dd["lons"] = self._swath_merger(
            "lons",
            selection=filled_selection,
            set_auto_mask=set_auto_mask,
            meter=False,
            fill_value=np.nan,
        )["lons"]
        dd["height"] = self._swath_merger(
            "height",
            selection=filled_selection,
            set_auto_mask=set_auto_mask,
            meter=False,
            fill_value=np.nan,
        )["height"]
        return dd

    def merge_correction(
        self,
        name: CorrectionType = ECorrectionType.SUM,
        selection=None,
        set_auto_mask: bool = True,
        meter: bool = False,
        direction: Literal["x", "y"] | None = None,
    ):
        """Merge multiple swaths of the specified correction variable.

        Data of the selected swaths (typically overlapped) are merged
        together to form a single data matrix with a consistent (range and
        azimuth) time axis.

        Note
        ----

        The current implementation uses a very simple algorithm that
        iterates over selected swaths and bursts and stitches correction
        data together.

        In overlapping regions, new data simply overwrite the old ones.
        This is an easy algorithm and perfectly correct for atmospheric
        and geodetic correction.

        It is, instead, sub-optimal for system corrections (bi-static,
        Doppler, FM Rate) which have different values in overlapping
        regions. In this case results are *not* correct.

        Parameters
        ----------
        name : str or ECorrectionType
            the name of the desired correction
        selection : list or pandas.DataFrame
            list of selected bursts (by default all bursts are selected)
        set_auto_mask : bool
            requested for netCDF4 to avoid retrieving a masked array
        meter : bool
            transform the result in meters
        direction : str or None
            if set to "x" (for range) or "y" (for "azimuth") only extracts
            the specified correction component.
            By default (None) all available components are returned.

        Returns
        -------
        dict
            a dictionary containing merged data and sampling information:

            :<burst_var_name>:
                merged data for the selected burst_var
            :first_azimuth_time:
                the relative azimuth first time
            :first_slant_range_time:
                the relative (slant) range first time
            :sampling:
                a dictionary containing the sampling along the
                'x' and 'y' directions and the 'unit'
            :units:
                of the correction (seconds or meters)
            :lats:
                the matrix of latitude values (in degrees) for each point
            :lons:
                the matrix of longitude values (in degrees) for each point
            :height:
                the matrix of height values (in meters) for each point

        """
        correction_type = ECorrectionType(name)  # check values
        prm_list = _CORRECTION_NAMES_MAP[correction_type.value]
        if direction is not None:
            prm_list = {direction: prm_list[direction]}
        correction = self._core_merge_correction(
            prm_list,
            selection=selection,
            set_auto_mask=set_auto_mask,
            meter=meter,
        )
        correction["name"] = correction_type.value
        return correction


class Sentinel1EtadSwath:
    """Object representing a swath in the S1-ETAD product.

    This objects are returned by methods of the :class:`Sentine1Etad` class.
    It is not expected that the user instantiates this objects directly.
    """

    def __init__(self, nc_group, parent):
        """Initialize a `Sentinel1EtadSwath` object."""
        self._grp = nc_group
        self._parent = weakref.proxy(parent)
        self._bursts = {}

    def __getitem__(self, burst_index):
        if burst_index not in self._bursts:
            burst_name = f"Burst{burst_index:04d}"
            group = self._grp[burst_name]
            self._bursts[burst_index] = Sentinel1EtadBurst(group, self)
        return self._bursts[burst_index]

    def __iter__(self):
        yield from self.iter_bursts()

    def __repr__(self):
        return f'{self.__class__.__name__}("{self._grp.path}")  0x{id(self):x}'

    @property
    def burst_list(self):
        """Return the list of burst identifiers of all bursts in the swath."""
        return [int(burst.bIndex) for burst in self._grp.groups.values()]

    @property
    def number_of_burst(self):
        """Return the number of bursts in the swath."""
        return len(self._grp.groups)

    @property
    def swath_id(self):
        """Return the swath identifier (str)."""
        return self._grp.swathID

    @property
    def swath_index(self):
        """Return the swath index (int)."""
        return int(self._grp.sIndex)

    @property
    def sampling_start(self):
        """Return the relative sampling start times."""
        first_burst_index = self.burst_list[0]
        first_burst = self[first_burst_index]
        return first_burst.sampling_start

    @property
    def sampling(self):
        """Return the sampling in seconds used for all bursts of the swath.

        A dictionary containing the following keys:

        * "x": range spacing,
        * "y": azimuth spacing,
        * "units": the measurement units used for "x' and "y"
        """
        first_burst_index = self.burst_list[0]
        first_burst = self[first_burst_index]
        return first_burst.sampling

    def _selection_to_burst_index_list(self, selection=None):
        if selection is None:
            index_list = self.burst_list
        elif isinstance(selection, pd.DataFrame):
            idx = selection.swathID == self.swath_id
            index_list = selection.bIndex[idx].values
        else:
            index_list = selection
        return index_list

    def iter_bursts(self, selection=None):
        """Iterate over bursts according to the specified selection.

        Parameters
        ----------
        selection : list(int) or pd.Dataframe, optional
            the list of selected bursts or result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the burst of the swath.

        """
        index_list = self._selection_to_burst_index_list(selection)
        for burst_index in index_list:
            yield self[burst_index]

    def get_footprint(self, selection=None, merge: bool = False):
        """Return the footprints of all the bursts as MultiPolygon.

        It calls in the back the get_footprint of the Sentinel1EtadBurst class.

        Parameters
        ----------
        selection : list(int) or pd.Dataframe, optional
            the list of selected bursts or result of a
            Sentinel1Etad.query_burst query.
            If the selection is None (default) the iteration is performed
            on all the burst of the swath.
        merge : bool
            if set to True return a single polygon that is the union of the
            footprints of all bursts

        """
        poly_list = [
            burst.get_footprint() for burst in self.iter_bursts(selection)
        ]
        if merge:
            polys = shapely.ops.unary_union(poly_list)
        else:
            polys = MultiPolygon(poly_list)

        return polys

    def intersects(self, geometry: BaseGeometry):
        """Return the list of burst indexes intersecting the input geometry.

        Computes the intersection of the footprint of the swath (all bursts)
        with the input Geometry

        Parameters
        ----------
        geometry : shapely.geometry.[Point, Polygon, MultiPolygon, line]
            shapely geometry

        Returns
        -------
        list
            list of the indexes of all bursts intersecting with the input
            geometry

        """
        if not isinstance(geometry, BaseGeometry):
            raise TypeError(
                "The input shape is not a shapely.BaseGeometry object"
            )
        burst_index_list = []
        swath_footprint = self.get_footprint(merge=True)
        if swath_footprint.intersects(geometry):
            burst_index_list = [
                b.burst_index
                for b in self.iter_bursts()
                if b.intersects(geometry)
            ]
        return burst_index_list

    def _burst_merger(
        self,
        burst_var,
        selection=None,
        az_time_min=None,
        az_time_max=None,
        set_auto_mask: bool = False,
        meter: bool = False,
        fill_value: float = 0.0,
    ):
        """Low level method to de-burst a NetCDF variable.

        The de-burst strategy is simple as the latest line is on top of the
        oldest.

        Parameters
        ----------
        burst_var : str
            one of the burst netcdf variables
        selection : list or pandas.DataFrame
            list of selected bursts (by default all bursts are selected)
        az_time_min : float | None
            minimum azimuth time of the merged swath
             (relative to the reference annotated in the NetCDF root)
        az_time_max : float | None
            maximum azimuth tim eof the merged swath
            (relative to the reference annotated in the NetCDF root)
        set_auto_mask : bool
            requested for netCDF4 to avoid retrieving a masked array
        meter : bool
            transform the result in meters
        fill_value: float
            fill value to be used in areas not covered by data (default: 0.0)

        Returns
        -------
        dict
            a dictionary containing merged data and sampling information:

            :<burst_var_name>: merged data for the selected burst_var
            :first_azimuth_time: the relative azimuth first time
            :first_slant_range_time: the relative (slant) range first time
            :sampling: a dictionary containing the sampling along the
            'x' and 'y' directions and the 'unit'

        """
        burst_index_list = self._selection_to_burst_index_list(selection)

        # Find what is the extent of the acquisition in azimuth
        first_burst = self[burst_index_list[0]]
        last_burst = self[burst_index_list[-1]]

        if az_time_min is None:
            t0 = first_burst.sampling_start["y"]
        else:
            t0 = az_time_min

        last_azimuth, _ = last_burst.get_burst_grid()
        if az_time_max is None:
            t1 = last_azimuth[-1]
        else:
            t1 = az_time_max

        tau0 = min(
            burst.sampling_start["x"]
            for burst in self.iter_bursts(burst_index_list)
        )

        # grid sampling
        dt = first_burst.sampling["y"]
        dtau = first_burst.sampling["x"]

        num_lines = np.round((t1 - t0) / dt).astype(int) + 1
        num_samples = max(
            burst.samples for burst in self.iter_bursts(burst_index_list)
        )

        debursted_var = np.full(
            (num_lines, num_samples), fill_value=fill_value
        )
        # TODO: add some control option
        debursted_var = np.ma.array(
            debursted_var, mask=True, fill_value=fill_value
        )

        for burst_ in self.iter_bursts(burst_index_list):
            if dt != burst_.sampling["y"]:
                raise RuntimeError(
                    "The azimuth sampling is changing long azimuth"
                )
            if first_burst.sampling_start["x"] != burst_.sampling_start["x"]:
                raise RuntimeError(
                    "The 2-way range gridStartRangeTime is changing "
                    "long azimuth"
                )

            # get the timing of the burst and convert into line index
            az_time_, rg_time_ = burst_.get_burst_grid()
            line_index_ = np.round((az_time_ - t0) / dt).astype(int)
            p0 = np.round((rg_time_[0] - tau0) / dtau).astype(int)

            # NOTE: use the private "Sentinel1EtadBurst._get_etad_param" method
            # to be able to work only on the specified NetCDF variable
            var_ = burst_._get_etad_param(
                burst_var,
                set_auto_mask=set_auto_mask,
                meter=meter,
            )

            _, burst_samples = var_.shape
            debursted_var[line_index_, p0 : p0 + burst_samples] = var_

        return {
            burst_var: debursted_var,
            "first_azimuth_time": t0,
            "first_slant_range_time": first_burst.sampling_start["x"],
            "sampling": first_burst.sampling,
        }

    def _core_merge_correction(
        self,
        prm_list,
        selection=None,
        set_auto_mask: bool = True,
        meter: bool = False,
    ):
        dd = {}
        for dim, field in prm_list.items():
            dd_ = self._burst_merger(
                field,
                selection=selection,
                set_auto_mask=set_auto_mask,
                meter=meter,
            )
            dd[dim] = dd_[field]
            dd["sampling"] = dd_["sampling"]
            dd["first_azimuth_time"] = dd_["first_azimuth_time"]
            dd["first_slant_range_time"] = dd_["first_slant_range_time"]

        dd["unit"] = "m" if meter else "s"
        dd["lats"] = self._burst_merger(
            "lats", set_auto_mask=set_auto_mask, meter=False
        )["lats"]
        dd["lons"] = self._burst_merger(
            "lons", set_auto_mask=set_auto_mask, meter=False
        )["lons"]
        dd["height"] = self._burst_merger(
            "height", set_auto_mask=set_auto_mask, meter=False
        )["height"]
        return dd

    def merge_correction(
        self,
        name: CorrectionType = ECorrectionType.SUM,
        selection=None,
        set_auto_mask: bool = True,
        meter: bool = False,
        direction: Literal["x", "y"] | None = None,
    ):
        """Merge multiple bursts of the specified correction variable.

        Data of the selected bursts (typically overlapped) are merged
        together to form a single data matrix with a consistent (azimuth)
        time axis.

        Note
        ----

        The current implementation uses a very simple algorithm that
        iterates over selected bursts and stitches correction data
        together.

        In overlapping regions, new data simply overwrite the old ones.
        This is an easy algorithm and perfectly correct for atmospheric
        and geodetic correction.

        It is, instead, sub-optimal for system corrections (bi-static,
        Doppler, FM Rate) which have different values in overlapping
        regions. In this case results are *not* correct.

        Parameters
        ----------
        name : str or CorrectionType
            the name of the desired correction
        selection : list or pandas.DataFrame
            list of selected bursts (by default all bursts are selected)
        set_auto_mask : bool
            requested for netCDF4 to avoid retrieving a masked array
        meter : bool
            transform the result in meters
        direction : str or None
            if set to "x" (for range) or "y" (for "azimuth") only extracts
            the specified correction component.
            By default (None) all available components are returned.

        Returns
        -------
        dict
            a dictionary containing merged data and sampling information:

            :<burst_var_name>:
                merged data for the selected burst_var
            :first_azimuth_time:
                the relative azimuth first time
            :first_slant_range_time:
                the relative (slant) range first time
            :sampling:
                a dictionary containing the sampling along the
                'x' and 'y' directions and the 'unit'
            :units:
                of the correction (seconds or meters)
            :lats:
                the matrix of latitude values (in degrees) for each point
            :lons:
                the matrix of longitude values (in degrees) for each point
            :height:
                the matrix of height values (in meters) for each point

        """
        correction_type = ECorrectionType(name)  # check values
        prm_list = _CORRECTION_NAMES_MAP[correction_type.value]
        if direction is not None:
            prm_list = {direction: prm_list[direction]}
        correction = self._core_merge_correction(
            prm_list,
            selection=selection,
            set_auto_mask=set_auto_mask,
            meter=meter,
        )
        correction["name"] = correction_type.value
        return correction


class Sentinel1EtadBurst:
    """Object representing a burst in the S1-ETAD product.

    This objects are returned by methods of the :class:`Sentinel1EtadSwath`
    class.
    It is not expected that the user instantiates this objects directly.
    """

    def __init__(self, nc_group, parent):
        """Initialize a `Sentinel1EtadBurst` object."""
        self._grp = nc_group
        self._parent = weakref.proxy(parent)
        self._geocoder = None
        self._footprint = None

    def __repr__(self):
        return f'{self.__class__.__name__}("{self._grp.path}")  0x{id(self):x}'

    @property
    def product_id(self) -> str:
        """Return the S1 product name (str) to which the burst belongs."""
        return self._grp.productID

    @property
    def swath_id(self) -> str:
        """Return the swath identifier (str) to which the burst belongs."""
        return self._grp.swathID

    @property
    def burst_id(self) -> str:
        """Return the burst identifier."""
        return self._grp.name

    @property
    def product_index(self) -> int:
        """Return the index of the S1 product to which the burst belongs."""
        return int(self._grp.pIndex)

    @property
    def swath_index(self) -> int:
        """Return the index (int) of the swath to which the burst belongs."""
        return int(self._grp.sIndex)

    @property
    def burst_index(self) -> int:
        """Return the index (int) of the burst."""
        return int(self._grp.bIndex)

    def get_footprint(self):
        """Return the footprint of the bursts as shapely.Polygon.

        It gets the lat/lon/height grid and extract the 4 corners.
        """
        if self._footprint is None:
            lats, lons, heights = self.get_lat_lon_height()
            corner_list = [(0, 0), (0, -1), (-1, -1), (-1, 0)]
            etad_burst_footprints = []
            for corner in corner_list:
                lat_, lon_, h_ = lats[corner], lons[corner], heights[corner]
                etad_burst_footprints.append((lon_, lat_, h_))
            self._footprint = Polygon(etad_burst_footprints)
        return self._footprint

    def intersects(self, geometry: BaseGeometry):
        """Intersect the footprint of the burst with the provided shape.

        Parameters
        ----------
        geometry : shapely.geometry.[Point, Polygon, MultiPolygon, line]
            shapely geometry

        Returns
        -------
        bool
            True if intersects, False otherwise

        """
        if not isinstance(geometry, BaseGeometry):
            raise TypeError("Not a shapely BaseGeometry object")
        return self.get_footprint().intersects(geometry)

    def get_burst_grid(self):
        """Return the t, tau grid of the burst."""
        azimuth = self._get_etad_param("azimuth", set_auto_mask=True)
        range_ = self._get_etad_param("range", set_auto_mask=True)
        return azimuth, range_

    @property
    def sampling_start(self):
        """Relative sampling start times.

        Value in seconds relative to the beginning of the product.
        """
        # TODO: put a reference in the docstring to the proper
        #       Sentinel1Etad property.
        return {
            "x": self._grp.gridStartRangeTime,
            "y": self._grp.gridStartAzimuthTime,
            "units": "s",
        }

    @property
    def sampling(self):
        """Return the sampling in seconds used for all bursts of the swath.

        A dictionary containing the following keys:

        * "x": range spacing,
        * "y": azimuth spacing,
        * "units": the measurement units used for "x' and "y"
        """
        return {
            "x": self._grp.gridSamplingRange,
            "y": self._grp.gridSamplingAzimuth,
            "units": "s",
        }

    @property
    def lines(self):
        """Return the number of lines in  the burst."""
        return self._grp.dimensions["azimuthExtent"].size

    @property
    def samples(self):
        """Return the number of samples in the burst."""
        return self._grp.dimensions["rangeExtent"].size

    @property
    def vg(self) -> float:
        """Average zero-Doppler ground velocity [m/s]."""
        return self._grp.averageZeroDopplerVelocity

    @property
    def reference_polarization(self) -> str:
        """Reverence polarization (string)."""
        try:
            return self._grp.referencePolarisation
        except AttributeError:
            return self._grp.referencePolarization

    def get_polarimetric_channel_offset(self, channel: str) -> dict:
        """Polarimetric channel delay.

        Return the electronic delay of the specified polarimetric channel
        w.r.t. the reference one (see
        :data:`Sentinel1EtadBurst.reference_polarization`).

        channel : str
            the string ID of the requested polarimetric channel:
            * 'VV' or 'VH' for DV products
            * 'HH' or 'HV' for DH products
        """
        if channel not in {"HH", "HV", "VV", "VH"}:
            raise ValueError(f"invalid channel ID: {channel!r}")

        if channel[0] != self.reference_polarization[0]:
            raise ValueError(
                f"polarimetric channel not available: {channel!r}"
            )

        data = {"units": "s"}

        if channel == "HH":
            data["x"] = self._grp.rangeOffsetHH
            data["y"] = self._grp.azimuthOffsetHH
        elif channel == "HV":
            data["x"] = self._grp.rangeOffsetHV
            data["y"] = self._grp.azimuthOffsetHV
        elif channel == "VH":
            data["x"] = self._grp.rangeOffsetVH
            data["y"] = self._grp.azimuthOffsetVH
        elif channel == "VV":
            data["x"] = self._grp.rangeOffsetVV
            data["y"] = self._grp.azimuthOffsetVV

        return data

    def get_timing_calibration_constants(self) -> dict:
        """Return the calibration constant for timing."""
        try:
            return {
                "x": self._grp.instrumentTimingCalibrationRange,
                "y": self._grp.instrumentTimingCalibrationAzimuth,
                "units": "s",
            }
        except AttributeError:
            # @COMPATIBILITY: with SETAP , v1.6
            warnings.warn(
                "instrument timing calibration constants are not available "
                "in the NetCDF data component this product. "
                "Calibration constants have been added to the NetCDF "
                "component in SETAP v1.6 (ETAD-DLR-PS-0014 - "
                '"ETAD Product Format Specification" Issue 1.5).',
                stacklevel=2,
            )
            return {"x": 0, "y": 0, "units": "s"}

    def _get_etad_param(
        self,
        name: str,
        set_auto_mask: bool = False,
        transpose: bool = False,
        meter: bool = False,
    ):
        if name not in self._grp.variables:
            raise RuntimeError(f"Parameter {name!r} is not allowed")

        self._grp.set_auto_mask(set_auto_mask)

        # TODO: avoid double copies
        # TODO: decimation factor
        field = np.asarray(self._grp[name])
        if transpose:
            field = np.transpose(field)

        if meter:
            if name.endswith("Az"):
                k = self._grp.averageZeroDopplerVelocity
            elif name.endswith("Rg"):
                k = constants.c / 2
            else:
                # it is not a correction (azimuth, range, lats, lons, height)
                k = 1
                warnings.warn(
                    f"the {name} is not a correction: "
                    "the 'meter' parameter will be ignored",
                    stacklevel=2,
                )
            field *= k

        return field

    def get_lat_lon_height(self, transpose: bool = False):
        """Return the latitude, longitude and height for each point.

        Data are returned as (3) matrices (lines x samples).
        Latitude and longitude are expressed in degrees, height is
        expressed in meters.
        """
        lats = self._get_etad_param(
            "lats", transpose=transpose, meter=False, set_auto_mask=True
        )
        lons = self._get_etad_param(
            "lons", transpose=transpose, meter=False, set_auto_mask=True
        )
        h = self._get_etad_param(
            "height", transpose=transpose, meter=False, set_auto_mask=True
        )
        return lats, lons, h

    def _core_get_correction(
        self,
        prm_list,
        set_auto_mask: bool = False,
        transpose: bool = False,
        meter: bool = False,
    ):
        correction = {}
        for dim, field in prm_list.items():
            correction[dim] = self._get_etad_param(
                field,
                set_auto_mask=set_auto_mask,
                transpose=transpose,
                meter=meter,
            )

        correction["unit"] = "m" if meter else "s"

        return correction

    def get_correction(
        self,
        name: CorrectionType = ECorrectionType.SUM,
        set_auto_mask: bool = False,
        transpose: bool = False,
        meter: bool = False,
        direction: Literal["x", "y"] | None = None,
    ):
        """Retrieve the correction for the specified correction "name".

        Puts the results in a dict.

        Parameters
        ----------
        name : ECorrectionType or str
            the desired correction
        set_auto_mask : bool
            requested for netCDF4 to avoid retrieving a masked array
        transpose : bool
            requested to retrieve the correction in array following the
            numpy convention for dimensions (default: False)
        meter : bool
            transform the result in meters
        direction : str or None
            if set to "x" (for range) or "y" (for "azimuth") only extracts
            the specified correction component.
            By default (None) all available components are returned.

        Returns
        -------
        dict
            a dictionary containing the following items for the
            requested correction:

            :x: correction in range (if applicable)
            :y: correction in azimuth (if applicable)
            :unit: 'm' or 's'
            :name: name of the correction

        """
        correction_type = ECorrectionType(name)  # check values
        processor_version = self._parent._parent._processor_version
        if processor_version < "003.00" and correction_type in {
            ECorrectionType.OTL,
            ECorrectionType.TROPOSPHERIC_GRADIENT,
        }:
            raise RuntimeError(
                f"correction {correction_type} is not available for "
                "processor version lower than '003.00' "
                f"(processor_version: {processor_version!r})"
            )

        name = correction_type.value
        prm_list = _CORRECTION_NAMES_MAP[name]
        if direction is not None:
            prm_list = {direction: prm_list[direction]}
        correction = self._core_get_correction(
            prm_list,
            set_auto_mask=set_auto_mask,
            transpose=transpose,
            meter=meter,
        )
        correction["name"] = name
        return correction

    def _get_geocoder(self):
        if self._geocoder is None:
            from .geometry import GridGeocoding

            azimuth, range_ = self.get_burst_grid()
            lats, lons, heights = self.get_lat_lon_height()
            self._geocoder = GridGeocoding(
                lats, lons, heights, xaxis=range_, yaxis=azimuth
            )
        return self._geocoder

    def radar_to_geodetic(self, tau, t, deg: bool = True):
        """Convert RADAR coordinates into geodetic coordinates.

        Compute the geodetic coordinates (lat, lon, h) corresponding to
        RADAR coordinates (tau, t), i.e. fast time (range time) and slow
        time (azimuth time expressed in seconds form the reference
        :data:`Sentinel1Etad.min_azimuth_time`)::

            (tau, t) -> (lat, lon, h)

        If ``deg`` is True the output ``lat`` and ``lon`` are expressed
        in degrees, otherwise in radians.

        The implementation is approximated and exploits pre-computed grids
        of latitude, longitude and height values.

        The method is not as accurate as solving the range-Doppler equations.

        .. seealso:: :class:`s1etad.geometry.GridGeocoding`.
        """
        return self._get_geocoder().forward_geocode(tau, t, deg=deg)

    def geodetic_to_radar(self, lat, lon, h=0, deg: bool = True):
        """Convert geodetic coordinates into RADAR coordinates.

        Compute the RADAR coordinates (tau, t), i.e. fast time (range time)
        and slow time (azimuth time expressed in seconds form the reference
        :data:`Sentinel1Etad.min_azimuth_time`) corresponding to
        geodetic coordinates (lat, lon, h)::

            (lat, lon, h) -> (tau, t)

        If ``deg`` is True it is assumed that input ``lat`` and ``lon``
        are expressed in degrees, otherwise it is assumed that angles
        are expressed in radians.

        The implementation is approximated and exploits pre-computed grids
        of latitude, longitude and height values.

        The method is not as accurate as solving the range-Doppler equations.

        .. seealso:: :class:`s1etad.geometry.GridGeocoding`.
        """
        return self._get_geocoder().backward_geocode(lat, lon, h, deg=deg)

    def radar_to_image(self, t, tau):
        """Convert RADAR coordinates into image coordinates.

        Compute the image coordinates (line, sample) corresponding
        to RADAR coordinates (tau, t), i.e. fast time (range time) and
        slow time (azimuth time expressed in seconds form the reference
        :data:`Sentinel1Etad.min_azimuth_time`)::

            (tau, t) -> (line, sample)
        """
        line = (t - self.sampling_start["y"]) / self.sampling["y"]
        sample = (tau - self.sampling_start["x"]) / self.sampling["x"]
        return line, sample

    def image_to_radar(self, line, sample):
        """Convert image coordinates into RADAR coordinates.

        Compute the RADAR coordinates (tau, t), i.e. fast time (range time)
        and slow time (azimuth time expressed in seconds form the reference
        :data:`Sentinel1Etad.min_azimuth_time`) corresponding to
        image coordinates (line, sample)::

            (line, sample) -> (t, tau)
        """
        t = self.sampling_start["y"] + line * self.sampling["y"]
        tau = self.sampling_start["x"] + sample * self.sampling["x"]
        return t, tau



"""Utility functions for the management of command line arguments."""

import logging
import argparse
import importlib
from collections.abc import Callable

from .. import __version__

DEFAULT_LOGLEVEL = "WARNING"


def _autocomplete(parser: argparse.ArgumentParser) -> None:
    try:
        import argcomplete
    except ImportError:
        pass
    else:
        argcomplete.autocomplete(parser)


def _add_logging_control_args(
    parser: argparse.ArgumentParser, default_loglevel: str = DEFAULT_LOGLEVEL
) -> argparse.ArgumentParser:
    """Add command line options for logging control."""
    loglevels = [logging.getLevelName(level) for level in range(10, 60, 10)]

    parser.add_argument(
        "--loglevel",
        default=default_loglevel,
        choices=loglevels,
        help="logging level (default: %(default)s)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        dest="loglevel",
        action="store_const",
        const="ERROR",
        help=(
            "suppress standard output messages, "
            "only errors are printed to screen (set 'loglevel' to 'ERROR')"
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="loglevel",
        action="store_const",
        const="INFO",
        help="print verbose output messages (set 'loglevel' to 'INFO')",
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="loglevel",
        action="store_const",
        const="DEBUG",
        help="print debug messages (set 'loglevel' to 'DEBUG')",
    )

    return parser


def finalize_parser(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Finalize the argument parser provided in input.

    Arguments for logging control and version printing are added.
    If the "argcomplete" module is available the automatic completion for
    the parser is also set-up.

    The updated parser is returned.
    """
    parser = _add_logging_control_args(parser)
    parser.add_argument(
        "--version", action="version", version="%(prog)s v" + __version__
    )

    _autocomplete(parser)

    return parser


def get_function(func: str | Callable) -> Callable:
    """Return the function corresponding to the input.

    If the input `func` parameter is already a callable then it is
    immediately returned.

    If the input is a string in the form "modname.[...].funcname" then the
    required Python package/module is imported and the specified function
    ("funcname" in the example) is returned.
    """
    if callable(func):
        return func

    fullname: str = func
    if "." not in fullname:
        raise ValueError(
            f"unable to retrieve the module name from: {fullname}"
        )

    modulename, funcname = fullname.rsplit(".", maxsplit=1)
    module = importlib.import_module(modulename)
    return getattr(module, funcname)



import re

import pytest
import shapely
from pytest_lazy_fixtures import lf as lazy_fixture

import s1etad


@pytest.mark.network
@pytest.mark.parametrize(
    "etad_swath",
    [
        pytest.param(lazy_fixture("sm_etad_swath"), id="SM"),
        pytest.param(lazy_fixture("iw_etad_swath"), id="IW"),
    ],
)
class TestSwath:
    @staticmethod
    def test_burst_list(etad_swath):
        burst_list = etad_swath.burst_list
        assert all(isinstance(burst_id, int) for burst_id in burst_list)
        assert len(burst_list) >= 0
        if etad_swath.swath_id.startswith("S"):
            assert len(burst_list) == 1

    @staticmethod
    def test_get_footprint(etad_swath):
        footprint = etad_swath.get_footprint()
        assert isinstance(footprint, shapely.MultiPolygon)
        for geom in footprint.geoms:
            assert geom.has_z

        footprint_no_merge = etad_swath.get_footprint(merge=False)
        assert isinstance(footprint_no_merge, shapely.MultiPolygon)
        for geom in footprint_no_merge.geoms:
            assert geom.has_z
        assert footprint == footprint_no_merge

    @staticmethod
    def test_get_footprint_merge(etad_swath):
        footprint = etad_swath.get_footprint(merge=True)
        assert isinstance(footprint, shapely.Polygon)
        assert footprint.has_z

    @staticmethod
    def test_intersects(etad_swath):
        assert etad_swath.intersects(etad_swath.get_footprint().centroid)
        assert not etad_swath.intersects(shapely.Point(0, 90, 0))

    @staticmethod
    def test_iter_bursts(etad_swath):
        bursts = list(etad_swath.iter_bursts())
        assert len(bursts) > 0
        if etad_swath.swath_id.startswith("S"):
            assert len(bursts) == 1
        assert all(
            isinstance(burst, s1etad.Sentinel1EtadBurst)
            for burst in etad_swath.iter_bursts()
        )

    @staticmethod
    def test_number_of_burst(etad_swath):
        if etad_swath.swath_id.startswith("S"):
            assert etad_swath.number_of_burst == 1
        else:
            assert etad_swath.number_of_burst > 1

    @staticmethod
    def test_sampling(etad_swath):
        sampling = etad_swath.sampling
        assert isinstance(sampling, dict)
        assert set(sampling.keys()) == {"x", "y", "units"}
        assert sampling["units"] == "s"

    @staticmethod
    def test_sampling_start(etad_swath):
        sampling_start = etad_swath.sampling_start
        assert isinstance(sampling_start, dict)
        assert set(sampling_start.keys()) == {"x", "y", "units"}
        assert sampling_start["units"] == "s"

    @staticmethod
    def test_swath_id(etad_swath):
        assert isinstance(etad_swath.swath_id, str)
        assert re.match(r"S\d|IW|EW", etad_swath.swath_id)

    @staticmethod
    def test_swath_index(etad_swath):
        assert isinstance(etad_swath.swath_index, int)
        assert etad_swath.swath_index > 0


"""
TODO:

    Sentinel1EtadSwath
        + get_footprint(selection)
        + iter_bursts(selection)
        merge_correction(name, selection, set_auto_mask, meter, direction)

        __getitem__
        __iter__
        __repr__
"""



from .product import Sentinel1Etad, Sentinel1EtadBurst, Sentinel1EtadSwath


def _sentinel1_etad_repr_pretty_(obj, p, cycle):
    if cycle:
        p.text(repr(obj))
    else:
        p.text(repr(obj))
        p.break_()
        plist = obj.s1_product_list()
        if isinstance(plist, str):
            plist = [plist]
        p.text(f"Number of Sentinel-1 slices: {len(plist)}")
        p.break_()
        with p.group(2, "Sentinel-1 products list:"):
            for name in plist:
                p.break_()
                p.text(name)
        p.break_()
        p.text(f"Number of swaths: {obj.number_of_swath}")
        p.break_()
        p.text("Swath list: {}".format(", ".join(obj.swath_list)))
        p.break_()
        with p.group(2, "Azimuth time:"):
            p.break_()
            p.text(f"min: {obj.min_azimuth_time}")
            p.break_()
            p.text(f"max: {obj.max_azimuth_time}")
        p.break_()
        with p.group(2, "Range time:"):
            p.break_()
            p.text(f"min: {obj.min_range_time}")
            p.break_()
            p.text(f"max: {obj.max_range_time}")
        p.break_()
        with p.group(2, "Grid sampling:"):
            for key, value in obj.grid_sampling.items():
                p.break_()
                p.text(f"{key}: {value}")
        p.break_()
        with p.group(2, "Grid spacing:"):
            for key, value in obj.grid_spacing.items():
                p.break_()
                p.text(f"{key}: {value}")
        p.break_()
        with p.group(2, "Processing settings:"):
            for key, value in obj.processing_setting().items():
                p.break_()
                p.text(f"{key}: {value}")


def _sentinel1_etad_swath_repr_pretty_(obj, p, cycle):
    if cycle:
        p.text(repr(obj))
    else:
        p.text(repr(obj))
        p.break_()
        p.text(f"Swaths ID: {obj.swath_id}")
        p.break_()
        p.text(f"Number of bursts: {obj.number_of_burst}")
        p.break_()
        p.text("Burst list: " + str(obj.burst_list))
        p.break_()
        with p.group(2, "Sampling start:"):
            for key, value in obj.sampling_start.items():
                p.break_()
                p.text(f"{key}: {value}")
        p.break_()
        with p.group(2, "Sampling:"):
            for key, value in obj.sampling.items():
                p.break_()
                p.text(f"{key}: {value}")


def _sentinel1_etad_burst_repr_pretty_(obj, p, cycle):
    if cycle:
        p.text(repr(obj))
    else:
        p.text(repr(obj))
        p.break_()
        p.text(f"Swaths ID: {obj.swath_id}")
        p.break_()
        p.text(f"Burst index: {obj.burst_index}")
        p.break_()
        p.text(f"Shape: ({obj.lines}, {obj.samples})")
        p.break_()
        with p.group(2, "Sampling start:"):
            for key, value in obj.sampling_start.items():
                p.break_()
                p.text(f"{key}: {value}")
        p.break_()
        with p.group(2, "Sampling:"):
            for key, value in obj.sampling.items():
                p.break_()
                p.text(f"{key}: {value}")


def _register_jupyter_formatters():
    try:
        from IPython import get_ipython
    except ImportError:
        pass
    else:
        ipy = get_ipython()

        if ipy is not None:
            formatter = ipy.display_formatter.formatters["text/plain"]
            formatter.for_type(Sentinel1Etad, _sentinel1_etad_repr_pretty_)
            formatter.for_type(
                Sentinel1EtadSwath, _sentinel1_etad_swath_repr_pretty_
            )
            formatter.for_type(
                Sentinel1EtadBurst, _sentinel1_etad_burst_repr_pretty_
            )
            return True
    return False



"""Utility functions for unit testing."""

import shutil
import pathlib

import pooch

BASEURL = "https://sentinels.copernicus.eu/documents/d/sentinel"
DATADIR = pathlib.Path(__file__).parent / "data"
REGISTRY = DATADIR.joinpath("registry.txt")


datarepo = pooch.create(
    path=DATADIR,
    base_url=BASEURL,
    # retry_if_failed=3,
    # version="0.5"
    registry=None,
    env="S1ETAD_TEST_DATADIR",
)
datarepo.load_registry(REGISTRY)


def download_all(datarepo: pooch.Pooch = datarepo):
    for item in datarepo.registry_files:
        datarepo.fetch(item, processor=pooch.Untar(), progressbar=True)


def clean_cache(datarepo: pooch.Pooch = datarepo):
    for item in datarepo.registry_files:
        shutil.rmtree(
            datarepo.path.joinpath(f"{item}.untar"), ignore_errors=True
        )
        datarepo.path.joinpath(item).unlink(missing_ok=True)



"""Unittest for the s1eats.utils module."""

import pytest

import s1etad.utils
from s1etad import ECorrectionType


@pytest.mark.parametrize(
    ("input_", "output"),
    [
        pytest.param(
            None,
            [
                (ECorrectionType.BISTATIC, "y"),
                (ECorrectionType.DOPPLER, "x"),
                (ECorrectionType.FMRATE, "y"),
                (ECorrectionType.GEODETIC, "x"),
                (ECorrectionType.GEODETIC, "y"),
                (ECorrectionType.IONOSPHERIC, "x"),
                (ECorrectionType.SUM, "x"),
                (ECorrectionType.SUM, "y"),
                (ECorrectionType.TROPOSPHERIC, "x"),
            ],
            id="default",
        ),
        pytest.param(
            [
                ECorrectionType.GEODETIC,
                ECorrectionType.IONOSPHERIC,
                ECorrectionType.TROPOSPHERIC,
            ],
            [
                (ECorrectionType.GEODETIC, "x"),
                (ECorrectionType.GEODETIC, "y"),
                (ECorrectionType.IONOSPHERIC, "x"),
                (ECorrectionType.TROPOSPHERIC, "x"),
            ],
            id="enums",
        ),
        pytest.param(
            ["geodetic", "ionospheric", "tropospheric"],
            [
                (ECorrectionType.GEODETIC, "x"),
                (ECorrectionType.GEODETIC, "y"),
                (ECorrectionType.IONOSPHERIC, "x"),
                (ECorrectionType.TROPOSPHERIC, "x"),
            ],
            id="strings",
        ),
    ],
)
def test_iter_corrections(input_, output):
    out = sorted(s1etad.utils.iter_corrections(input_), key=str)
    assert out == output



"""Generate a KMZ file containing previews of the S1-ETAD corrections."""

import argparse

from . import utils as cliutils


def get_parser(subparsers=None) -> argparse.ArgumentParser:
    """Instantiate the command line argument parser."""
    name = "export-kmz"
    synopsis = __doc__.splitlines()[0]
    doc = __doc__

    if subparsers is None:
        parser = argparse.ArgumentParser(prog=name, description=doc)
    else:
        parser = subparsers.add_parser(name, description=doc, help=synopsis)

    parser.set_defaults(func="s1etad.kmz.etad_to_kmz")

    # command line options
    # ...

    # positional arguments
    parser.add_argument(
        "etad",
        metavar="etad-path",
        help="path to the S1-ETAD products (a directory)",
    )
    parser.add_argument(
        "outpath",
        nargs="?",
        help=(
            "the pathname of the generated KMZ file. "
            "If not provided then a KMZ file having the same input product "
            "basename is generated in the current working directory"
        ),
    )

    if not subparsers:
        parser = cliutils.finalize_parser(parser)

    return parser



"""Simple CLI tool to access S1-ETAD products.

The tool provides a set of sub-commands to perform basic tasks involving
S1-ETAD products.
"""

# PYTHON_ARGCOMPLETE_OK

import logging
import argparse
from collections.abc import Sequence

from . import exportkmz, ql
from . import utils as cliutils

EX_OK = 0
EX_FAILURE = 1
EX_INTERRUPT = 130

PROG = __package__.split(".")[0]
LOGFMT = "%(asctime)s %(name)s %(levelname)s -- %(message)s"
DEFAULT_LOGLEVEL = "INFO"


def get_parser() -> argparse.ArgumentParser:
    """Instantiate the command line argument parser."""
    description = __doc__
    parser = argparse.ArgumentParser(prog=PROG, description=description)

    # Sub-command management
    subparsers = parser.add_subparsers(title="sub-commands")  # dest='func'
    exportkmz.get_parser(subparsers=subparsers)
    ql.get_parser(subparsers=subparsers)

    parser = cliutils.finalize_parser(parser)

    return parser


def parse_args(
    args: Sequence[str] | None = None,
    namespace: argparse.Namespace | None = None,
    parser: argparse.ArgumentParser | None = None,
) -> argparse.Namespace:
    """Parse command line arguments."""
    if parser is None:
        parser = get_parser()

    out_args = parser.parse_args(args, namespace)

    # Common pre-processing of parsed arguments and consistency checks
    # ...

    if getattr(out_args, "func", None) is None:
        parser.error("no sub-command specified.")

    return out_args


def _get_kwargs(args) -> dict:
    kwargs = vars(args).copy()
    kwargs.pop("func", None)
    kwargs.pop("loglevel", None)
    return kwargs


def main(*argv: str) -> int:
    """Implement the main CLI interface."""
    # setup logging
    logging.basicConfig(format=LOGFMT, level=DEFAULT_LOGLEVEL)
    logging.captureWarnings(True)
    log = logging.getLogger(PROG)

    # parse cmd line arguments
    args = parse_args(argv if argv else None)

    exit_code = EX_OK
    try:
        # NOTE: use the root logger to set the logging level
        logging.getLogger().setLevel(args.loglevel)

        log.debug("args: %s", args)

        func = cliutils.get_function(args.func)
        kwargs = _get_kwargs(args)
        func(**kwargs)
    except Exception as exc:  # noqa: B902, BLE001
        log.critical(
            "unexpected exception caught: %r %s", type(exc).__name__, exc
        )
        log.debug("stacktrace:", exc_info=True)
        exit_code = EX_FAILURE
    except KeyboardInterrupt:
        log.warning("Keyboard interrupt received: exit the program")
        exit_code = EX_INTERRUPT

    return exit_code



"""Unit tests for ETAD products."""

import re
import pathlib
import datetime

import pandas as pd
import pytest
import netCDF4
import shapely
from pytest_lazy_fixtures import lf as lazy_fixture

import s1etad
from s1etad import ECorrectionType

COLUMNS = (
    "bIndex",
    "pIndex",
    "sIndex",
    "productID",
    "swathID",
    "azimuthTimeMin",
    "azimuthTimeMax",
)


CORRECTION_SETTINGS = {
    "troposphericDelayCorrection",
    "troposphericDelayCorrectionGradient",
    "ionosphericDelayCorrection",
    "solidEarthTideCorrection",
    "oceanTidalLoadingCorrection",
    "bistaticAzimuthCorrection",
    "dopplerShiftRangeCorrection",
    "FMMismatchAzimuthCorrection",
}


def _get_mode_id(path):
    name = pathlib.Path(path).name
    mobj = re.match(r"S1[AB]_(?P<mode_id>[SIE][W0-9])_ETA_.*\.SAFE", name)
    assert mobj is not None
    mode_id = mobj.group("mode_id")
    return "SM" if mode_id.startswith("S") else mode_id


def _get_n_swaths(mode_id):
    if mode_id == "SM":
        n_swaths = 1
    elif mode_id == "IW":
        n_swaths = 3
    elif mode_id == "EW":
        n_swaths = 5
    else:
        raise ValueError(f"invalid mode_id: {mode_id}")
    return n_swaths


@pytest.mark.network
@pytest.mark.parametrize(
    "filename",
    [
        pytest.param(lazy_fixture("sm_etad_filename"), id="SM"),
        pytest.param(lazy_fixture("iw_etad_filename"), id="IW"),
    ],
)
def test_open_etad_product(filename):
    eta = s1etad.Sentinel1Etad(filename)
    assert eta


@pytest.mark.network
@pytest.mark.parametrize(
    "etad_product",
    [
        pytest.param(lazy_fixture("sm_etad_product"), id="SM"),
        pytest.param(lazy_fixture("iw_etad_product"), id="IW"),
    ],
)
class TestEtadProduct:
    @staticmethod
    def test_processor_version(etad_product):
        assert isinstance(etad_product._processor_version, str)
        assert re.match(r"\d{3}\.\d{2}", etad_product._processor_version)

    @staticmethod
    def test_burst_catalogue(etad_product):
        burst_catalogue = etad_product.burst_catalogue
        assert isinstance(burst_catalogue, pd.DataFrame)
        assert len(burst_catalogue) > 0
        if _get_mode_id(etad_product.product) == "SM":
            assert len(burst_catalogue) == 1
        assert tuple(burst_catalogue.columns) == COLUMNS

    @staticmethod
    def test_ds(etad_product):
        ds = etad_product.ds
        assert isinstance(ds, netCDF4.Dataset)

    @staticmethod
    def test_get_footprint(etad_product):
        footprint = etad_product.get_footprint()
        assert isinstance(footprint, shapely.MultiPolygon)
        for geom in footprint.geoms:
            assert geom.has_z

        footprint_no_merge = etad_product.get_footprint(merge=False)
        assert isinstance(footprint_no_merge, shapely.MultiPolygon)
        for geom in footprint_no_merge.geoms:
            assert geom.has_z
        assert footprint == footprint_no_merge

    @staticmethod
    def test_get_footprint_merge(etad_product):
        footprint = etad_product.get_footprint(merge=True)
        assert isinstance(footprint, shapely.Polygon)
        assert footprint.has_z

    @staticmethod
    @pytest.mark.parametrize(
        "correction_type",
        [
            "tropospheric",
            "tropospheric_gradient",
            "ionospheric",
            "geodetic",
            "otl",
            "bistatic",
            "doppler",
            "fmrate",
            "sum",
        ],
    )
    def test_get_statistics(etad_product, correction_type):
        if _get_mode_id(etad_product.product) == "SM" and correction_type in {
            "doppler",
            "fmrate",
        }:
            pytest.xfail(reason="layer not available for SM")

        if etad_product._processor_version < "003.00" and correction_type in {
            "tropospheric_gradient",
            "otl",
        }:
            pytest.xfail(
                reason=(
                    f"correction {correction_type!r} not available for "
                    "processor version smaller than '003.00' "
                    f"(processot_version: {etad_product._processor_version!r})"
                )
            )

        stats = etad_product.get_statistics(correction_type)
        assert set(stats.keys()).issubset({"x", "y", "unit"})

        correction = ECorrectionType(correction_type)
        if correction is ECorrectionType.TROPOSPHERIC_GRADIENT:
            assert stats["unit"] == "s/m"
        else:
            assert stats["unit"] == "s"

    @staticmethod
    def test_get_statistics_units(etad_product):
        assert etad_product.get_statistics("sum", meter=True)["unit"] == "m"

    @staticmethod
    def test_grid_sampling(etad_product):
        grid_sampling = etad_product.grid_sampling
        assert isinstance(grid_sampling, dict)
        assert set(grid_sampling.keys()) == {"x", "y", "unit"}
        assert grid_sampling["unit"] == "s"

    @staticmethod
    def test_grid_spacing(etad_product):
        grid_spacing = etad_product.grid_spacing
        assert isinstance(grid_spacing, dict)
        assert set(grid_spacing.keys()) == {"x", "y", "unit"}
        assert grid_spacing["unit"] == "m"

    @staticmethod
    def test_intersects(etad_product):
        assert etad_product.intersects(etad_product.get_footprint().centroid)
        assert not etad_product.intersects(shapely.Point(0, 90, 0))

    @staticmethod
    def test_iter_bursts(etad_product):
        bursts = list(etad_product.iter_bursts())
        assert len(bursts) > 0
        if _get_mode_id(etad_product.product) == "SM":
            assert len(bursts) == 1
        assert all(
            isinstance(burst, s1etad.Sentinel1EtadBurst)
            for burst in etad_product.iter_bursts()
        )

    @staticmethod
    def test_iter_swaths(etad_product):
        swaths = list(etad_product.iter_swaths())
        assert len(swaths) == etad_product.number_of_swath
        assert all(
            isinstance(swath, s1etad.Sentinel1EtadSwath)
            for swath in etad_product.iter_swaths()
        )

    @staticmethod
    def test_azimuth_time(etad_product):
        assert isinstance(etad_product.min_azimuth_time, datetime.datetime)
        assert isinstance(etad_product.max_azimuth_time, datetime.datetime)
        assert etad_product.min_azimuth_time <= etad_product.max_azimuth_time

    @staticmethod
    def test_range_time(etad_product):
        assert isinstance(etad_product.min_range_time, float)
        assert isinstance(etad_product.max_range_time, float)
        assert etad_product.min_range_time <= etad_product.max_range_time

    @staticmethod
    def test_number_of_swaths(etad_product):
        n_swaths = _get_n_swaths(_get_mode_id(etad_product.product))
        assert etad_product.number_of_swath == n_swaths

    @staticmethod
    def test_processing_setting(etad_product):
        settings = CORRECTION_SETTINGS.copy()
        if etad_product._processor_version < "003.00":
            settings.pop("oceanTidalLoadingCorrection")
        assert set(etad_product.processing_setting()) == settings

    @staticmethod
    def test_s1_etad_product_list(etad_product):
        assert len(etad_product.s1_product_list()) == 1
        assert all(
            re.match(r"S1[AB]_([SA]\d|IW|EW)_SLC_\w*\.SAFE", etad_product)
            for etad_product in etad_product.s1_product_list()
        )
        assert sorted(etad_product.ds.groups.keys()) == sorted(
            etad_product.swath_list
        )

    @staticmethod
    def test_swath_list(etad_product):
        mode_id = _get_mode_id(etad_product.product)
        if mode_id.startswith("S"):
            pattern = r"S\d"
        else:
            pattern = mode_id
        assert all(
            re.match(pattern, swath_id) for swath_id in etad_product.swath_list
        )

    @staticmethod
    def test_vg(etad_product):
        assert isinstance(etad_product.vg, float)
        assert 6000 <= etad_product.vg <= 8000


@pytest.mark.network
@pytest.mark.parametrize(
    ("etad_product", "etad_filename"),
    [
        (lazy_fixture("sm_etad_product"), lazy_fixture("sm_etad_filename")),
        (lazy_fixture("iw_etad_product"), lazy_fixture("iw_etad_filename")),
    ],
)
def test_product(etad_product, etad_filename):
    assert isinstance(etad_product.product, pathlib.Path)
    assert etad_product.product == pathlib.Path(etad_filename)


"""
TODO:

    Sentinel1Etad
        merge_correction(name, selection, set_auto_mask, meter, direction)
        query_burst(first_time, product_name, last_time, swath, geometry)
        __getitem__
        __iter__
        __repr__
        __str__
        + iter_bursts(selection)
        + iter_swaths(selection)
        + get_footprint(selection)
"""



"""Utility functions."""

from .product import ECorrectionType


def iter_corrections(corrections=None):
    """Iterate over corrections."""
    if corrections is None:
        corrections = ECorrectionType

    if isinstance(corrections, str) or isinstance(
        corrections, ECorrectionType
    ):
        corrections = [corrections]

    for correction in corrections:
        correction = ECorrectionType(correction)
        if correction in {
            ECorrectionType.TROPOSPHERIC,
            ECorrectionType.IONOSPHERIC,
            ECorrectionType.DOPPLER,
        }:
            yield correction, "x"
        elif correction in {ECorrectionType.BISTATIC, ECorrectionType.FMRATE}:
            yield correction, "y"
        elif correction in {ECorrectionType.SUM, ECorrectionType.GEODETIC}:
            yield correction, "x"
            yield correction, "y"



"""Geo-coded QuickLook image generation for ETAD."""

import os
import functools
from typing import Literal
from collections.abc import Sequence

import numpy as np
from osgeo import gdal, osr

from . import ECorrectionType, Sentinel1Etad
from .product import CorrectionType

MAX_GCP_NUM: int = 10000  # empirical threshold
# 0.005 [deg] --> 550m @ equator, 50 @ lat=85deg
DEFAULT_LATLON_SPACING_DEG: float = 0.005  # deg
DEFAULT_COLOR_TABLE_NAME: str = "jet"  # form matplotlib


def _write_band_data(band, data, nodata: float = -9999.0):
    if hasattr(data, "filled"):
        data = data.filled(nodata)
    band.WriteArray(data)
    band.SetNoDataValue(nodata)


def create_gcps(
    lat,
    lon,
    h=None,
    gcp_step: tuple[int, int] = (10, 10),
) -> list[gdal.GCP]:
    """Generate a sub-sampled grid of GCPs form input coordinate matrices."""
    assert lat.shape == lon.shape
    ysize, xsize = lat.shape
    ystep, xstep = gcp_step

    masks = [
        data.mask if hasattr(data, "mask") else None for data in (lat, lon, h)
    ]

    mask = None
    if masks:
        mask = functools.reduce(np.logical_or, masks)

    gcps: list[gdal.GCP] = []
    for line in range(0, ysize, ystep):
        for pix in range(0, xsize, xstep):
            if mask is None or mask[line, pix]:
                continue
            height = h[line, pix] if h is not None else 0.0
            gcp_info = ""
            gcp_id = f"{len(gcps)}"
            gcp = gdal.GCP(
                lon[line, pix],
                lat[line, pix],
                height,
                pix,
                line,
                gcp_info,
                gcp_id,
            )
            gcps.append(gcp)

    assert 0 < len(gcps) <= MAX_GCP_NUM

    return gcps


def save_with_gcps(
    outfile: str,
    data,
    lat,
    lon,
    h=None,
    *,
    drv_name: str = "GTIFF",
    nodata: float = -9999.0,
    gcp_step: tuple[int, int] = (10, 10),
    srs: str | osr.SpatialReference = "wgs84",
    creation_options: Sequence[str] | None = None,
):
    """Save data into a GDAL dataset and GCPs for coordinates matrices."""
    drv = gdal.GetDriverByName(drv_name)
    assert drv is not None

    ysize, xsize = data.shape
    if creation_options is None:
        creation_options = []
    ds = drv.Create(
        str(outfile),
        xsize=xsize,
        ysize=ysize,
        bands=1,
        eType=gdal.GDT_Float32,
        options=creation_options,
    )

    if isinstance(srs, str):
        srs_str = srs
        srs = osr.SpatialReference()
        srs.SetWellKnownGeogCS(srs_str)
    gcps = create_gcps(lat, lon, h, gcp_step)
    ds.SetGCPs(gcps, srs)

    _write_band_data(ds.GetRasterBand(1), data, nodata)

    return ds


def _clip_bbox(bbox, q, margin=0):
    return (
        np.floor(bbox[0] / q) * q - margin * q,
        np.floor(bbox[1] / q) * q - margin * q,
        np.ceil(bbox[2] / q) * q + margin * q,
        np.ceil(bbox[3] / q) * q + margin * q,
    )


def _compute_gcp_spacing(
    xsize: int, ysize: int, max_gcp_num: int = MAX_GCP_NUM
) -> tuple[int, int]:
    # assume 200 x 200 m ground spacing for ETAD products
    gcp_step = (25, 25)  # 5 x 5 km
    # gcp_step = (50, 50)  # 10 x 10 km
    # gcp_step = (100, 100)  # 20 x 20 km
    while (ysize // gcp_step[0]) * (xsize // gcp_step[1]) > max_gcp_num:
        # increase the step only in the azimuth direction
        gcp_step = (gcp_step[0] * 2, gcp_step[1])
    return gcp_step


@functools.lru_cache
def _get_color_table(name: str = DEFAULT_COLOR_TABLE_NAME):
    from matplotlib import cm

    from .kmz import Colorizer  # noqa: F401

    cmap = getattr(cm, name)
    # return Colorizer(1, 255, color_table=cmap).gdal_palette()

    table = gdal.ColorTable()
    table.SetColorEntry(0, (0, 0, 0, 0))  # zero is transparent
    for i, v in enumerate(np.linspace(0.0, 1.0, 255), start=1):
        table.SetColorEntry(i, cmap(v, bytes=True))
    return table


def save_geocoded_data(
    outfile,
    data,
    lat,
    lon,
    h=None,
    *,
    gcp_step: tuple[int, int] | None = None,
    srs: str | osr.SpatialReference = "wgs84",
    out_spacing: float = DEFAULT_LATLON_SPACING_DEG,
    drv_name="GTIFF",
    creation_options: Sequence[str] | None = None,
    palette: str = DEFAULT_COLOR_TABLE_NAME,
    margin=100,
):
    """Save a geo-coded version of input data into a GDAL dataset."""
    ysize, xsize = data.shape
    if gcp_step is None:
        gcp_step = _compute_gcp_spacing(xsize, ysize)

    # dataset with GCP grid
    ds_sr_with_gcps = save_with_gcps(
        "", data, lat=lat, lon=lon, h=h, gcp_step=gcp_step, drv_name="MEM"
    )

    # geocode the floating point image
    bbox = (lon.min(), lat.min(), lon.max(), lat.max())
    bbox = _clip_bbox(bbox, out_spacing, margin=margin)

    ds_geocoded_float = gdal.Warp(
        "",
        ds_sr_with_gcps,
        format="MEM",
        dstSRS=srs,
        xRes=out_spacing,
        yRes=out_spacing,
        targetAlignedPixels=True,
        outputBounds=bbox,
        outputBoundsSRS=srs,
    )

    # scale the geocoded image to bytes
    scale_params = [[data.min(), data.max(), 0, 255]]  # NOTE: list of lists
    ds_geocoded_bytes = gdal.Translate(
        "",
        ds_geocoded_float,
        format="MEM",
        outputType=gdal.GDT_Byte,
        noData=0,
        scaleParams=scale_params,
    )
    # attach the color palette
    if isinstance(palette, str):
        palette = _get_color_table()

    band = ds_geocoded_bytes.GetRasterBand(1)
    band.SetRasterColorTable(palette)
    band.SetRasterColorInterpretation(gdal.GCI_PaletteIndex)
    del band

    # Save to disk
    if creation_options is None:
        creation_options = []
    ds_out = gdal.Translate(
        os.fspath(outfile),
        ds_geocoded_bytes,
        format=drv_name,
        creationOptions=creation_options,
    )
    # , rgbExpand='rgba')

    return ds_out


def etad2ql(
    etad,
    outpath,
    *,
    correction_type: CorrectionType = ECorrectionType.SUM,
    direction: Literal["x", "y"] = "x",
    meter: bool = True,
    drv_name: str = "PNG",
    creation_options: Sequence[str] = ("WORLDFILE=YES",),
):
    """Generate a geo-coded quick-look image starting from an ETAD product."""
    if not isinstance(etad, Sentinel1Etad):
        etad = Sentinel1Etad(etad)
    else:
        etad = etad

    if outpath is None:
        outpath = etad.product.with_suffix(".png").name

    merged_correction = etad.merge_correction(correction_type, meter=meter)

    # masked arrays
    data = merged_correction[direction]
    lat = merged_correction["lats"]
    lon = merged_correction["lons"]
    height = merged_correction["height"]

    return save_geocoded_data(
        outpath,
        data,
        lat,
        lon,
        height,
        drv_name=drv_name,
        creation_options=list(creation_options),
    )



import numpy as np
import pytest
import shapely
from pytest_lazy_fixtures import lf as lazy_fixture

import s1etad


@pytest.mark.network
@pytest.mark.parametrize(
    "etad_burst",
    [
        pytest.param(lazy_fixture("sm_etad_burst"), id="SM"),
        pytest.param(lazy_fixture("iw_etad_burst"), id="IW"),
    ],
)
class TestBurst:
    @staticmethod
    def test_burst_id(etad_burst):
        assert isinstance(etad_burst.burst_id, str)
        assert etad_burst.burst_id.startswith("Burst")

    @staticmethod
    def test_burst_index(etad_burst):
        assert isinstance(etad_burst.burst_index, int)
        assert etad_burst.burst_index > 0

    @staticmethod
    def test_get_burst_grid(etad_burst):
        az, rg = etad_burst.get_burst_grid()
        assert isinstance(az, np.ndarray)
        assert isinstance(rg, np.ndarray)
        assert az.dtype == np.float64
        assert rg.dtype == np.float64
        assert len(az) == etad_burst.lines
        assert len(rg) == etad_burst.samples

    @staticmethod
    def test_get_footprint(etad_burst):
        footprint = etad_burst.get_footprint()
        assert isinstance(footprint, shapely.Polygon)
        assert footprint.has_z

    @staticmethod
    @pytest.mark.parametrize("channel", ["H", "V"])
    def test_get_polarimetric_channel_offset(etad_burst, channel):
        channel = f"{etad_burst.reference_polarization[0]}{channel}"
        offset = etad_burst.get_polarimetric_channel_offset(channel)
        assert set(offset.keys()) == {"x", "y", "units"}
        assert offset["units"] == "s"
        assert isinstance(offset["x"], float)
        assert isinstance(offset["y"], float)

    @staticmethod
    def test_get_polarimetric_channel_offset_invalid_channel(etad_burst):
        with pytest.raises(ValueError):
            etad_burst.get_polarimetric_channel_offset("invalid")

    @staticmethod
    def test_get_timing_calibration_constants(etad_burst):
        k = etad_burst.get_timing_calibration_constants()
        assert set(k.keys()) == {"x", "y", "units"}
        assert k["units"] == "s"
        assert isinstance(k["x"], float)
        assert isinstance(k["y"], float)

    @staticmethod
    def test_reference_polarization(etad_burst):
        rp = etad_burst.reference_polarization
        assert isinstance(rp, str)
        assert rp in {"HH", "HV", "VH", "VV"}

    @staticmethod
    def test_lines(etad_burst):
        assert isinstance(etad_burst.lines, int)
        assert etad_burst.lines > 0

    @staticmethod
    def test_product_id(etad_burst):
        assert isinstance(etad_burst.product_id, str)
        assert etad_burst.product_id.startswith("S1")

    @staticmethod
    def test_product_index(etad_burst):
        assert isinstance(etad_burst.product_index, int)
        assert etad_burst.product_index >= 0

    @staticmethod
    def test_samples(etad_burst):
        assert isinstance(etad_burst.samples, int)
        assert etad_burst.samples > 0

    @staticmethod
    def test_sampling(etad_burst):
        s = etad_burst.sampling
        assert set(s.keys()) == {"x", "y", "units"}
        assert s["units"] == "s"
        assert isinstance(s["x"], float)
        assert isinstance(s["y"], float)
        assert s["x"] >= 0
        assert s["y"] >= 0

    @staticmethod
    def test_sampling_start(etad_burst):
        sstart = etad_burst.sampling_start
        assert set(sstart.keys()) == {"x", "y", "units"}
        assert sstart["units"] == "s"
        assert isinstance(sstart["x"], float)
        assert isinstance(sstart["y"], float)
        assert sstart["x"] >= 0
        assert sstart["y"] >= 0

    @staticmethod
    def test_swath_id(etad_burst):
        assert isinstance(etad_burst.swath_id, str)
        assert 2 <= len(etad_burst.swath_id) <= 3
        assert etad_burst.swath_id[0] in {"E", "I", "S", "W"}

    @staticmethod
    def test_swath_index(etad_burst):
        assert isinstance(etad_burst.swath_index, int)
        assert etad_burst.swath_index > 0

    @staticmethod
    def test_vg(etad_burst):
        assert isinstance(etad_burst.vg, float)
        assert etad_burst.vg > 0

    @staticmethod
    @pytest.mark.parametrize("name", list(s1etad.ECorrectionType))
    def test_get_correction_name(etad_burst, name):
        data = etad_burst.get_correction(name)
        assert data["name"] == name.value
        assert data["unit"] == "s"
        assert "x" in data or "y" in data
        if "x" in data:
            assert isinstance(data["x"], np.ndarray)
            assert data["x"].dtype == np.float64
        if "y" in data:
            assert isinstance(data["y"], np.ndarray)
            assert data["y"].dtype == np.float64

    @staticmethod
    @pytest.mark.parametrize(
        "name",
        [
            s1etad.ECorrectionType.TROPOSPHERIC_GRADIENT,
            s1etad.ECorrectionType.OTL,
        ],
    )
    def test_get_correction_name_pre_3(etad_burst, name, monkeypatch):
        with monkeypatch.context() as m:
            etad_product = etad_burst._parent._parent
            m.setattr(etad_product, "_processor_version", "002.92")
            with pytest.raises(RuntimeError, match="processor_version"):
                etad_burst.get_correction(name)


"""
TODO:

    Sentinel1EtadBurst
        get_correction(name, set_auto_mask, transpose, meter, direction)
        get_lat_lon_height(transpose)
        geodetic_to_radar(lat, lon, h, deg)
        image_to_radar
        intersects(geometry)
        radar_to_geodetic(tau, t, deg)
        __repr__
"""
